# -*- coding: utf-8 -*-
from mahjong.tile import TilesConverter


class Meld(object):
    CHI = 'chi'
    PON = 'pon'
    KAN = 'kan'
    CHANKAN = 'chankan'
    NUKI = 'nuki'

    who = None
    tiles = None
    type = None
    from_who = None
    called_tile = None
    # we need it to distinguish opened and closed kan
    opened = True

    def __init__(self, meld_type=None, tiles=None, opened=True, called_tile=None, who=None, from_who=None):
        self.type = meld_type
        self.tiles = tiles or []
        self.opened = opened
        self.called_tile = called_tile
        self.who = who
        self.from_who = from_who

    def __str__(self):
        return 'Type: {}, Tiles: {} {}'.format(self.type, TilesConverter.to_one_line_string(self.tiles), self.tiles)

    # for calls in array
    def __repr__(self):
        return self.__str__()

    @property
    def tiles_34(self):
        return [x // 4 for x in self.tiles[:3]]



# -*- coding: utf-8 -*-
from mahjong.constants import CHUN
from mahjong.hand_calculating.yaku import Yaku
from mahjong.utils import is_pon


class Chun(Yaku):
    """
    Pon of red dragons
    """

    def __init__(self, yaku_id=None):
        super(Chun, self).__init__(yaku_id)

    def set_attributes(self):
        self.tenhou_id = 20

        self.name = 'Yakuhai (chun)'
        self.english = 'Red Dragon'
        self.japanese = '役牌(中)'

        self.han_open = 1
        self.han_closed = 1

        self.is_yakuman = False

    def is_condition_met(self, hand, *args):
        return len([x for x in hand if is_pon(x) and x[0] == CHUN]) == 1



# 1 and 9
TERMINAL_INDICES = [0, 8, 9, 17, 18, 26]

# dragons and winds
EAST = 27
SOUTH = 28
WEST = 29
NORTH = 30
HAKU = 31
HATSU = 32
CHUN = 33

WINDS = [EAST, SOUTH, WEST, NORTH]
HONOR_INDICES = WINDS + [HAKU, HATSU, CHUN]

FIVE_RED_MAN = 16
FIVE_RED_PIN = 52
FIVE_RED_SOU = 88

AKA_DORA_LIST = [FIVE_RED_MAN, FIVE_RED_PIN, FIVE_RED_SOU]

DISPLAY_WINDS = {
    EAST: 'East',
    SOUTH: 'South',
    WEST: 'West',
    NORTH: 'North'
}



# -*- coding: utf-8 -*-

from mahjong.constants import TERMINAL_INDICES, HONOR_INDICES
from mahjong.meld import Meld
from mahjong.utils import is_pair, is_pon, simplify, contains_terminals


class FuCalculator(object):
    BASE = 'base'
    PENCHAN = 'penchan'
    KANCHAN = 'kanchan'
    VALUED_PAIR = 'valued_pair'
    PAIR_WAIT = 'pair_wait'
    TSUMO = 'tsumo'
    HAND_WITHOUT_FU = 'hand_without_fu'

    CLOSED_PON = 'closed_pon'
    OPEN_PON = 'open_pon'

    CLOSED_TERMINAL_PON = 'closed_terminal_pon'
    OPEN_TERMINAL_PON = 'open_terminal_pon'

    CLOSED_KAN = 'closed_kan'
    OPEN_KAN = 'open_kan'

    CLOSED_TERMINAL_KAN = 'closed_terminal_kan'
    OPEN_TERMINAL_KAN = 'open_terminal_kan'

    def calculate_fu(self,
                     hand,
                     win_tile,
                     win_group,
                     config,
                     valued_tiles=None,
                     melds=None,):
        """
        Calculate hand fu with explanations
        :param hand:
        :param win_tile: 136 tile format
        :param win_group: one set where win tile exists
        :param is_tsumo:
        :param config: HandConfig object
        :param valued_tiles: dragons, player wind, round wind
        :param melds: opened sets
        :return:
        """

        win_tile_34 = win_tile // 4

        if not valued_tiles:
            valued_tiles = []

        if not melds:
            melds = []

        fu_details = []

        if len(hand) == 7:
            return [{'fu': 25, 'reason': FuCalculator.BASE}], 25

        pair = [x for x in hand if is_pair(x)][0]
        pon_sets = [x for x in hand if is_pon(x)]

        copied_opened_melds = [x.tiles_34 for x in melds if x.type == Meld.CHI]
        closed_chi_sets = []
        for x in hand:
            if x not in copied_opened_melds:
                closed_chi_sets.append(x)
            else:
                copied_opened_melds.remove(x)

        is_open_hand = any([x.opened for x in melds])

        if win_group in closed_chi_sets:
            tile_index = simplify(win_tile_34)

            # penchan
            if contains_terminals(win_group):
                # 1-2-... wait
                if tile_index == 2 and win_group.index(win_tile_34) == 2:
                    fu_details.append({'fu': 2, 'reason': FuCalculator.PENCHAN})
                # 8-9-... wait
                elif tile_index == 6 and win_group.index(win_tile_34) == 0:
                    fu_details.append({'fu': 2, 'reason': FuCalculator.PENCHAN})

            # kanchan waiting 5-...-7
            if win_group.index(win_tile_34) == 1:
                fu_details.append({'fu': 2, 'reason': FuCalculator.KANCHAN})

        # valued pair
        count_of_valued_pairs = valued_tiles.count(pair[0])
        if count_of_valued_pairs == 1:
            fu_details.append({'fu': 2, 'reason': FuCalculator.VALUED_PAIR})

        # east-east pair when you are on east gave double fu
        if count_of_valued_pairs == 2:
            fu_details.append({'fu': 2, 'reason': FuCalculator.VALUED_PAIR})
            fu_details.append({'fu': 2, 'reason': FuCalculator.VALUED_PAIR})

        # pair wait
        if is_pair(win_group):
            fu_details.append({'fu': 2, 'reason': FuCalculator.PAIR_WAIT})

        for set_item in pon_sets:
            open_meld = [x for x in melds if set_item == x.tiles_34]
            open_meld = open_meld and open_meld[0] or None

            set_was_open = open_meld and open_meld.opened or False
            is_kan = (open_meld and (open_meld.type == Meld.KAN or open_meld.type == Meld.CHANKAN)) or False
            is_honor = set_item[0] in TERMINAL_INDICES + HONOR_INDICES

            # we win by ron on the third pon tile, our pon will be count as open
            if not config.is_tsumo and set_item == win_group:
                set_was_open = True

            if is_honor:
                if is_kan:
                    if set_was_open:
                        fu_details.append({'fu': 16, 'reason': FuCalculator.OPEN_TERMINAL_KAN})
                    else:
                        fu_details.append({'fu': 32, 'reason': FuCalculator.CLOSED_TERMINAL_KAN})
                else:
                    if set_was_open:
                        fu_details.append({'fu': 4, 'reason': FuCalculator.OPEN_TERMINAL_PON})
                    else:
                        fu_details.append({'fu': 8, 'reason': FuCalculator.CLOSED_TERMINAL_PON})
            else:
                if is_kan:
                    if set_was_open:
                        fu_details.append({'fu': 8, 'reason': FuCalculator.OPEN_KAN})
                    else:
                        fu_details.append({'fu': 16, 'reason': FuCalculator.CLOSED_KAN})
                else:
                    if set_was_open:
                        fu_details.append({'fu': 2, 'reason': FuCalculator.OPEN_PON})
                    else:
                        fu_details.append({'fu': 4, 'reason': FuCalculator.CLOSED_PON})

        add_tsumo_fu = len(fu_details) > 0 or config.options.fu_for_pinfu_tsumo

        if config.is_tsumo and add_tsumo_fu:
            # 2 additional fu for tsumo (but not for pinfu)
            fu_details.append({'fu': 2, 'reason': FuCalculator.TSUMO})

        if is_open_hand and not len(fu_details) and config.options.fu_for_open_pinfu:
            # there is no 1-20 hands, so we had to add additional fu
            fu_details.append({'fu': 2, 'reason': FuCalculator.HAND_WITHOUT_FU})

        if is_open_hand or config.is_tsumo:
            fu_details.append({'fu': 20, 'reason': FuCalculator.BASE})
        else:
            fu_details.append({'fu': 30, 'reason': FuCalculator.BASE})

        return fu_details, self.round_fu(fu_details)

    def round_fu(self, fu_details):
        # 22 -> 30 and etc.
        fu = sum([x['fu'] for x in fu_details])
        return (fu + 9) // 10 * 10



# -*- coding: utf-8 -*-
from mahjong.constants import HONOR_INDICES
from mahjong.hand_calculating.yaku import Yaku
from mahjong.utils import is_sou, is_pin, is_man


class Chinitsu(Yaku):
    """
    The hand contains tiles only from a single suit
    """

    def __init__(self, yaku_id=None):
        super(Chinitsu, self).__init__(yaku_id)

    def set_attributes(self):
        self.tenhou_id = 35

        self.name = 'Chinitsu'
        self.english = 'Flush'
        self.japanese = '清一色'

        self.han_open = 5
        self.han_closed = 6

        self.is_yakuman = False

    def is_condition_met(self, hand, *args):
        honor_sets = 0
        sou_sets = 0
        pin_sets = 0
        man_sets = 0
        for item in hand:
            if item[0] in HONOR_INDICES:
                honor_sets += 1

            if is_sou(item[0]):
                sou_sets += 1
            elif is_pin(item[0]):
                pin_sets += 1
            elif is_man(item[0]):
                man_sets += 1

        sets = [sou_sets, pin_sets, man_sets]
        only_one_suit = len([x for x in sets if x != 0]) == 1

        return only_one_suit and honor_sets == 0



# -*- coding: utf-8 -*-
from mahjong.hand_calculating.hand_config import HandConfig


class ScoresCalculator(object):

    def calculate_scores(self, han, fu, config, is_yakuman=False):
        """
        Calculate how much scores cost a hand with given han and fu
        :param han: int
        :param fu: int
        :param config: HandConfig object
        :param is_yakuman: boolean
        :return: a dictionary with main and additional cost
        for ron additional cost is always = 0
        for tsumo main cost is cost for dealer and additional is cost for player
        {'main': 1000, 'additional': 0}
        """

        # kazoe hand
        if han >= 13 and not is_yakuman:
            # Hands over 26+ han don't count as double yakuman
            if config.options.kazoe_limit == HandConfig.KAZOE_LIMITED:
                han = 13
            # Hands over 13+ is a sanbaiman
            elif config.options.kazoe_limit == HandConfig.KAZOE_SANBAIMAN:
                han = 12

        if han >= 5:
            if han >= 78:
                rounded = 48000
            elif han >= 65:
                rounded = 40000
            elif han >= 52:
                rounded = 32000
            elif han >= 39:
                rounded = 24000
            # double yakuman
            elif han >= 26:
                rounded = 16000
            # yakuman
            elif han >= 13:
                rounded = 8000
            # sanbaiman
            elif han >= 11:
                rounded = 6000
            # baiman
            elif han >= 8:
                rounded = 4000
            # haneman
            elif han >= 6:
                rounded = 3000
            else:
                rounded = 2000

            double_rounded = rounded * 2
            four_rounded = double_rounded * 2
            six_rounded = double_rounded * 3
        else:
            base_points = fu * pow(2, 2 + han)
            rounded = (base_points + 99) // 100 * 100
            double_rounded = (2 * base_points + 99) // 100 * 100
            four_rounded = (4 * base_points + 99) // 100 * 100
            six_rounded = (6 * base_points + 99) // 100 * 100

            is_kiriage = False
            if config.options.kiriage:
                if han == 4 and fu == 30:
                    is_kiriage = True
                if han == 3 and fu == 60:
                    is_kiriage = True

            # mangan
            if rounded > 2000 or is_kiriage:
                rounded = 2000
                double_rounded = rounded * 2
                four_rounded = double_rounded * 2
                six_rounded = double_rounded * 3

        if config.is_tsumo:
            return {'main': double_rounded, 'additional': config.is_dealer and double_rounded or rounded}
        else:
            return {'main': config.is_dealer and six_rounded or four_rounded, 'additional': 0}



# -*- coding: utf-8 -*-
from mahjong.hand_calculating.yaku import Yaku
from mahjong.utils import is_chi


class Iipeiko(Yaku):
    """
    Hand with two identical chi
    """

    def __init__(self, yaku_id=None):
        super(Iipeiko, self).__init__(yaku_id)

    def set_attributes(self):
        self.tenhou_id = 9

        self.name = 'Iipeiko'
        self.english = 'Identical Sequences'
        self.japanese = '一盃口'

        self.han_open = None
        self.han_closed = 1

        self.is_yakuman = False

    def is_condition_met(self, hand, *args):
        chi_sets = [i for i in hand if is_chi(i)]

        count_of_identical_chi = 0
        for x in chi_sets:
            count = 0
            for y in chi_sets:
                if x == y:
                    count += 1
            if count > count_of_identical_chi:
                count_of_identical_chi = count

        return count_of_identical_chi >= 2



# -*- coding: utf-8 -*-
from mahjong.hand_calculating.yaku import Yaku


class YakuhaiOfPlace(Yaku):

    def __init__(self, yaku_id=None):
        super(YakuhaiOfPlace, self).__init__(yaku_id)

    def set_attributes(self):
        self.tenhou_id = 10

        self.name = 'Yakuhai (wind of place)'
        self.english = 'Value Tiles (Seat)'
        self.japanese = '自風'

        self.han_open = 1
        self.han_closed = 1

        self.is_yakuman = False

    def is_condition_met(self, hand, *args):
        return True



# -*- coding: utf-8 -*-
from mahjong.hand_calculating.yaku import Yaku
from mahjong.meld import Meld


class SanKantsu(Yaku):
    """
    The hand with three kan sets
    """

    def __init__(self, yaku_id=None):
        super(SanKantsu, self).__init__(yaku_id)

    def set_attributes(self):
        self.tenhou_id = 27

        self.name = 'San Kantsu'
        self.english = 'Three Kans'
        self.japanese = '三槓子'

        self.han_open = 2
        self.han_closed = 2

        self.is_yakuman = False

    def is_condition_met(self, hand, melds, *args):
        kan_sets = [x for x in melds if x.type == Meld.KAN or x.type == Meld.CHANKAN]
        return len(kan_sets) == 3



from mahjong.hand_calculating.yaku_list.aka_dora import AkaDora
from mahjong.hand_calculating.yaku_list.chankan import Chankan
from mahjong.hand_calculating.yaku_list.chanta import Chanta
from mahjong.hand_calculating.yaku_list.chiitoitsu import Chiitoitsu
from mahjong.hand_calculating.yaku_list.chinitsu import Chinitsu
from mahjong.hand_calculating.yaku_list.chun import Chun
from mahjong.hand_calculating.yaku_list.daburu_riichi import DaburuRiichi
from mahjong.hand_calculating.yaku_list.dora import Dora
from mahjong.hand_calculating.yaku_list.haitei import Haitei
from mahjong.hand_calculating.yaku_list.haku import Haku
from mahjong.hand_calculating.yaku_list.hatsu import Hatsu
from mahjong.hand_calculating.yaku_list.honitsu import Honitsu
from mahjong.hand_calculating.yaku_list.honroto import Honroto
from mahjong.hand_calculating.yaku_list.houtei import Houtei
from mahjong.hand_calculating.yaku_list.iipeiko import Iipeiko
from mahjong.hand_calculating.yaku_list.ippatsu import Ippatsu
from mahjong.hand_calculating.yaku_list.ittsu import Ittsu
from mahjong.hand_calculating.yaku_list.junchan import Junchan
from mahjong.hand_calculating.yaku_list.nagashi_mangan import NagashiMangan
from mahjong.hand_calculating.yaku_list.pinfu import Pinfu
from mahjong.hand_calculating.yaku_list.renhou import Renhou
from mahjong.hand_calculating.yaku_list.riichi import Riichi
from mahjong.hand_calculating.yaku_list.rinshan import Rinshan
from mahjong.hand_calculating.yaku_list.ryanpeiko import Ryanpeikou
from mahjong.hand_calculating.yaku_list.sanankou import Sanankou
from mahjong.hand_calculating.yaku_list.sankantsu import SanKantsu
from mahjong.hand_calculating.yaku_list.sanshoku import Sanshoku
from mahjong.hand_calculating.yaku_list.sanshoku_douko import SanshokuDoukou
from mahjong.hand_calculating.yaku_list.shosangen import Shosangen
from mahjong.hand_calculating.yaku_list.tanyao import Tanyao
from mahjong.hand_calculating.yaku_list.toitoi import Toitoi
from mahjong.hand_calculating.yaku_list.tsumo import Tsumo
from mahjong.hand_calculating.yaku_list.yakuhai_place import YakuhaiOfPlace
from mahjong.hand_calculating.yaku_list.yakuhai_round import YakuhaiOfRound
from mahjong.hand_calculating.yaku_list.east import YakuhaiEast
from mahjong.hand_calculating.yaku_list.west import YakuhaiWest
from mahjong.hand_calculating.yaku_list.north import YakuhaiNorth
from mahjong.hand_calculating.yaku_list.south import YakuhaiSouth



# -*- coding: utf-8 -*-
from functools import reduce

from mahjong.hand_calculating.yaku import Yaku
from mahjong.utils import is_sou, is_pin, is_man, simplify


class ChuurenPoutou(Yaku):
    """
    The hand contains 1-1-1-2-3-4-5-6-7-8-9-9-9 of one suit, plus any other tile of the same suit.
    """

    def __init__(self, yaku_id=None):
        super(ChuurenPoutou, self).__init__(yaku_id)

    def set_attributes(self):
        self.tenhou_id = 45

        self.name = 'Chuuren Poutou'
        self.english = 'Nine Gates'
        self.japanese = '九蓮宝燈'

        self.han_open = None
        self.han_closed = 13

        self.is_yakuman = True

    def is_condition_met(self, hand, *args):
        sou_sets = 0
        pin_sets = 0
        man_sets = 0
        honor_sets = 0
        for item in hand:
            if is_sou(item[0]):
                sou_sets += 1
            elif is_pin(item[0]):
                pin_sets += 1
            elif is_man(item[0]):
                man_sets += 1
            else:
                honor_sets += 1

        sets = [sou_sets, pin_sets, man_sets]
        only_one_suit = len([x for x in sets if x != 0]) == 1
        if not only_one_suit or honor_sets > 0:
            return False

        indices = reduce(lambda z, y: z + y, hand)
        # cast tile indices to 0..8 representation
        indices = [simplify(x) for x in indices]

        # 1-1-1
        if len([x for x in indices if x == 0]) < 3:
            return False

        # 9-9-9
        if len([x for x in indices if x == 8]) < 3:
            return False

        # 1-2-3-4-5-6-7-8-9 and one tile to any of them
        indices.remove(0)
        indices.remove(0)
        indices.remove(8)
        indices.remove(8)
        for x in range(0, 9):
            if x in indices:
                indices.remove(x)

        if len(indices) == 1:
            return True

        return False



# -*- coding: utf-8 -*-
from mahjong.constants import EAST
from mahjong.hand_calculating.yaku_config import YakuConfig


class HandConstants(object):
    # Hands over 26+ han don't count as double yakuman
    KAZOE_LIMITED = 0
    # Hands over 13+ is a sanbaiman
    KAZOE_SANBAIMAN = 1
    # 26+ han as double yakuman, 39+ han as triple yakuman, etc.
    KAZOE_NO_LIMIT = 2


class OptionalRules(object):
    """
    All the supported optional rules
    """
    has_open_tanyao = False
    has_aka_dora = False
    has_double_yakuman = True
    kazoe_limit = HandConstants.KAZOE_LIMITED
    kiriage = False
    # if false, 1-20 hand will be possible
    fu_for_open_pinfu = True
    # if true, pinfu tsumo will be disabled
    fu_for_pinfu_tsumo = False
    renhou_as_yakuman = False
    has_daisharin = False
    has_daisharin_other_suits = False

    def __init__(self, has_open_tanyao=False,
                 has_aka_dora=False,
                 has_double_yakuman=True,
                 kazoe_limit=HandConstants.KAZOE_LIMITED,
                 kiriage=False,
                 fu_for_open_pinfu=True,
                 fu_for_pinfu_tsumo=False,
                 renhou_as_yakuman=False,
                 has_daisharin=False,
                 has_daisharin_other_suits=False):

        self.has_open_tanyao = has_open_tanyao
        self.has_aka_dora = has_aka_dora
        self.has_double_yakuman = has_double_yakuman
        self.kazoe_limit = kazoe_limit
        self.kiriage = kiriage
        self.fu_for_open_pinfu = fu_for_open_pinfu
        self.fu_for_pinfu_tsumo = fu_for_pinfu_tsumo
        self.renhou_as_yakuman = renhou_as_yakuman
        self.has_daisharin = has_daisharin or has_daisharin_other_suits
        self.has_daisharin_other_suits = has_daisharin_other_suits


class HandConfig(HandConstants):
    """
    Special class to pass various settings to the hand calculator object
    """
    yaku = None
    options = None

    is_tsumo = False
    is_riichi = False
    is_ippatsu = False
    is_rinshan = False
    is_chankan = False
    is_haitei = False
    is_houtei = False
    is_daburu_riichi = False
    is_nagashi_mangan = False
    is_tenhou = False
    is_renhou = False
    is_chiihou = False

    is_dealer = False
    player_wind = None
    round_wind = None

    def __init__(self,
                 is_tsumo=False,
                 is_riichi=False,
                 is_ippatsu=False,
                 is_rinshan=False,
                 is_chankan=False,
                 is_haitei=False,
                 is_houtei=False,
                 is_daburu_riichi=False,
                 is_nagashi_mangan=False,
                 is_tenhou=False,
                 is_renhou=False,
                 is_chiihou=False,
                 player_wind=None,
                 round_wind=None,
                 options=None):

        self.yaku = YakuConfig()
        self.options = options or OptionalRules()

        self.is_tsumo = is_tsumo
        self.is_riichi = is_riichi
        self.is_ippatsu = is_ippatsu
        self.is_rinshan = is_rinshan
        self.is_chankan = is_chankan
        self.is_haitei = is_haitei
        self.is_houtei = is_houtei
        self.is_daburu_riichi = is_daburu_riichi
        self.is_nagashi_mangan = is_nagashi_mangan
        self.is_tenhou = is_tenhou
        self.is_renhou = is_renhou
        self.is_chiihou = is_chiihou

        self.player_wind = player_wind
        self.round_wind = round_wind
        self.is_dealer = player_wind == EAST



# -*- coding: utf-8 -*-
from mahjong.hand_calculating.yaku import Yaku


class Chiitoitsu(Yaku):
    """
    Hand contains only pairs
    """

    def __init__(self, yaku_id=None):
        super(Chiitoitsu, self).__init__(yaku_id)

    def set_attributes(self):
        self.tenhou_id = 22

        self.name = 'Chiitoitsu'
        self.english = 'Seven Pairs'
        self.japanese = '七対子'

        self.han_open = None
        self.han_closed = 2

        self.is_yakuman = False

    def is_condition_met(self, hand, *args):
        return len(hand) == 7



# -*- coding: utf-8 -*-
from mahjong.constants import CHUN, HAKU, HATSU
from mahjong.hand_calculating.yaku import Yaku
from mahjong.utils import is_pon, is_pair


class Shosangen(Yaku):
    """
    Hand with two dragon pon sets and one dragon pair
    """

    def __init__(self, yaku_id=None):
        super(Shosangen, self).__init__(yaku_id)

    def set_attributes(self):
        self.tenhou_id = 30

        self.name = 'Shou Sangen'
        self.english = 'Small Three Dragons'
        self.japanese = '小三元'

        self.han_open = 2
        self.han_closed = 2

        self.is_yakuman = False

    def is_condition_met(self, hand, *args):
        dragons = [CHUN, HAKU, HATSU]
        count_of_conditions = 0
        for item in hand:
            # dragon pon or pair
            if (is_pair(item) or is_pon(item)) and item[0] in dragons:
                count_of_conditions += 1

        return count_of_conditions == 3



# -*- coding: utf-8 -*-
from mahjong.hand_calculating.yaku import Yaku
from mahjong.utils import is_chi, is_sou, is_pin, is_man, simplify


class Ittsu(Yaku):
    """
    Three sets of same suit: 1-2-3, 4-5-6, 7-8-9
    """

    def __init__(self, yaku_id=None):
        super(Ittsu, self).__init__(yaku_id)

    def set_attributes(self):
        self.tenhou_id = 24

        self.name = 'Ittsu'
        self.english = 'Straight'
        self.japanese = '一気通貫'

        self.han_open = 1
        self.han_closed = 2

        self.is_yakuman = False

    def is_condition_met(self, hand, *args):
        chi_sets = [i for i in hand if is_chi(i)]
        if len(chi_sets) < 3:
            return False

        sou_chi = []
        pin_chi = []
        man_chi = []
        for item in chi_sets:
            if is_sou(item[0]):
                sou_chi.append(item)
            elif is_pin(item[0]):
                pin_chi.append(item)
            elif is_man(item[0]):
                man_chi.append(item)

        sets = [sou_chi, pin_chi, man_chi]

        for suit_item in sets:
            if len(suit_item) < 3:
                continue

            casted_sets = []

            for set_item in suit_item:
                # cast tiles indices to 0..8 representation
                casted_sets.append([simplify(set_item[0]),
                                    simplify(set_item[1]),
                                    simplify(set_item[2])])

            if [0, 1, 2] in casted_sets and [3, 4, 5] in casted_sets and [6, 7, 8] in casted_sets:
                return True

        return False



# -*- coding: utf-8 -*-
from mahjong.constants import EAST, SOUTH, WEST, NORTH
from mahjong.hand_calculating.yaku import Yaku
from mahjong.utils import is_pon


class DaiSuushii(Yaku):
    """
    The hand contains four sets of winds
    """

    def __init__(self, yaku_id=None):
        super(DaiSuushii, self).__init__(yaku_id)

    def set_attributes(self):
        self.tenhou_id = 49

        self.name = 'Dai Suushii'
        self.english = 'Big Four Winds'
        self.japanese = '大四喜'

        self.han_open = 26
        self.han_closed = 26

        self.is_yakuman = True

    def is_condition_met(self, hand, *args):
        """
        The hand contains four sets of winds
        :param hand: list of hand's sets
        :return: boolean
        """
        pon_sets = [x for x in hand if is_pon(x)]
        if len(pon_sets) != 4:
            return False

        count_wind_sets = 0
        winds = [EAST, SOUTH, WEST, NORTH]
        for item in pon_sets:
            if is_pon(item) and item[0] in winds:
                count_wind_sets += 1

        return count_wind_sets == 4



# -*- coding: utf-8 -*-
from functools import reduce

from mahjong.constants import HONOR_INDICES
from mahjong.hand_calculating.yaku import Yaku


class Tsuuiisou(Yaku):
    """
    Hand composed entirely of honour tiles
    """

    def __init__(self, yaku_id=None):
        super(Tsuuiisou, self).__init__(yaku_id)

    def set_attributes(self):
        self.tenhou_id = 42

        self.name = 'Tsuu iisou'
        self.english = 'All Honors'
        self.japanese = '字一色'

        self.han_open = 13
        self.han_closed = 13

        self.is_yakuman = True

    def is_condition_met(self, hand, *args):
        """
        Hand composed entirely of honour tiles.
        :param hand: list of hand's sets
        :return: boolean
        """
        indices = reduce(lambda z, y: z + y, hand)
        return all(x in HONOR_INDICES for x in indices)



# -*- coding: utf-8 -*-
import copy

from mahjong.utils import find_isolated_tile_indices


class Agari(object):

    def is_agari(self, tiles_34, open_sets_34=None):
        """
        Determine was it win or not
        :param tiles_34: 34 tiles format array
        :param open_sets_34: array of array of 34 tiles format
        :return: boolean
        """
        # we will modify them later, so we need to use a copy
        tiles = copy.deepcopy(tiles_34)

        # With open hand we need to remove open sets from hand and replace them with isolated pon sets
        # it will allow to determine agari state correctly
        if open_sets_34:
            isolated_tiles = find_isolated_tile_indices(tiles)
            for meld in open_sets_34:
                if not isolated_tiles:
                    break

                isolated_tile = isolated_tiles.pop()

                tiles[meld[0]] -= 1
                tiles[meld[1]] -= 1
                tiles[meld[2]] -= 1
                tiles[isolated_tile] = 3

        j = (1 << tiles[27]) | (1 << tiles[28]) | (1 << tiles[29]) | (1 << tiles[30]) | \
            (1 << tiles[31]) | (1 << tiles[32]) | (1 << tiles[33])

        if j >= 0x10:
            return False

        # 13 orphans
        if ((j & 3) == 2) and (tiles[0] * tiles[8] * tiles[9] * tiles[17] * tiles[18] *
                               tiles[26] * tiles[27] * tiles[28] * tiles[29] * tiles[30] *
                               tiles[31] * tiles[32] * tiles[33] == 2):
            return True

        # seven pairs
        if not (j & 10) and sum([tiles[i] == 2 for i in range(0, 34)]) == 7:
            return True

        if j & 2:
            return False

        n00 = tiles[0] + tiles[3] + tiles[6]
        n01 = tiles[1] + tiles[4] + tiles[7]
        n02 = tiles[2] + tiles[5] + tiles[8]

        n10 = tiles[9] + tiles[12] + tiles[15]
        n11 = tiles[10] + tiles[13] + tiles[16]
        n12 = tiles[11] + tiles[14] + tiles[17]

        n20 = tiles[18] + tiles[21] + tiles[24]
        n21 = tiles[19] + tiles[22] + tiles[25]
        n22 = tiles[20] + tiles[23] + tiles[26]

        n0 = (n00 + n01 + n02) % 3
        if n0 == 1:
            return False

        n1 = (n10 + n11 + n12) % 3
        if n1 == 1:
            return False

        n2 = (n20 + n21 + n22) % 3
        if n2 == 1:
            return False

        if ((n0 == 2) + (n1 == 2) + (n2 == 2) + (tiles[27] == 2) + (tiles[28] == 2) +
                (tiles[29] == 2) + (tiles[30] == 2) + (tiles[31] == 2) + (tiles[32] == 2) +
                (tiles[33] == 2) != 1):
            return False

        nn0 = (n00 * 1 + n01 * 2) % 3
        m0 = self._to_meld(tiles, 0)
        nn1 = (n10 * 1 + n11 * 2) % 3
        m1 = self._to_meld(tiles, 9)
        nn2 = (n20 * 1 + n21 * 2) % 3
        m2 = self._to_meld(tiles, 18)

        if j & 4:
            return not (n0 | nn0 | n1 | nn1 | n2 | nn2) and self._is_mentsu(m0) \
                and self._is_mentsu(m1) and self._is_mentsu(m2)

        if n0 == 2:
            return not (n1 | nn1 | n2 | nn2) and self._is_mentsu(m1) and self._is_mentsu(m2) \
                and self._is_atama_mentsu(nn0, m0)

        if n1 == 2:
            return not (n2 | nn2 | n0 | nn0) and self._is_mentsu(m2) and self._is_mentsu(m0) \
                and self._is_atama_mentsu(nn1, m1)

        if n2 == 2:
            return not (n0 | nn0 | n1 | nn1) and self._is_mentsu(m0) and self._is_mentsu(m1) \
                and self._is_atama_mentsu(nn2, m2)

        return False

    def _is_mentsu(self, m):
        a = m & 7
        b = 0
        c = 0
        if a == 1 or a == 4:
            b = c = 1
        elif a == 2:
            b = c = 2
        m >>= 3
        a = (m & 7) - b

        if a < 0:
            return False

        is_not_mentsu = False
        for x in range(0, 6):
            b = c
            c = 0
            if a == 1 or a == 4:
                b += 1
                c += 1
            elif a == 2:
                b += 2
                c += 2
            m >>= 3
            a = (m & 7) - b
            if a < 0:
                is_not_mentsu = True
                break

        if is_not_mentsu:
            return False

        m >>= 3
        a = (m & 7) - c

        return a == 0 or a == 3

    def _is_atama_mentsu(self, nn, m):
        if nn == 0:
            if (m & (7 << 6)) >= (2 << 6) and self._is_mentsu(m - (2 << 6)):
                return True
            if (m & (7 << 15)) >= (2 << 15) and self._is_mentsu(m - (2 << 15)):
                return True
            if (m & (7 << 24)) >= (2 << 24) and self._is_mentsu(m - (2 << 24)):
                return True
        elif nn == 1:
            if (m & (7 << 3)) >= (2 << 3) and self._is_mentsu(m - (2 << 3)):
                return True
            if (m & (7 << 12)) >= (2 << 12) and self._is_mentsu(m - (2 << 12)):
                return True
            if (m & (7 << 21)) >= (2 << 21) and self._is_mentsu(m - (2 << 21)):
                return True
        elif nn == 2:
            if (m & (7 << 0)) >= (2 << 0) and self._is_mentsu(m - (2 << 0)):
                return True
            if (m & (7 << 9)) >= (2 << 9) and self._is_mentsu(m - (2 << 9)):
                return True
            if (m & (7 << 18)) >= (2 << 18) and self._is_mentsu(m - (2 << 18)):
                return True
        return False

    def _to_meld(self, tiles, d):
        result = 0
        for i in range(0, 9):
            result |= (tiles[d + i] << i * 3)
        return result



# -*- coding: utf-8 -*-
from mahjong.constants import CHUN, HAKU, HATSU
from mahjong.hand_calculating.yaku import Yaku
from mahjong.utils import is_pon


class Daisangen(Yaku):
    """
    The hand contains three sets of dragons
    """

    def __init__(self, yaku_id=None):
        super(Daisangen, self).__init__(yaku_id)

    def set_attributes(self):
        self.tenhou_id = 39

        self.name = 'Daisangen'
        self.english = 'Big Three Dragons'
        self.japanese = '大三元'

        self.han_open = 13
        self.han_closed = 13

        self.is_yakuman = True

    def is_condition_met(self, hand, *args):
        count_of_dragon_pon_sets = 0
        for item in hand:
            if is_pon(item) and item[0] in [CHUN, HAKU, HATSU]:
                count_of_dragon_pon_sets += 1
        return count_of_dragon_pon_sets == 3



# -*- coding: utf-8 -*-
from mahjong.constants import TERMINAL_INDICES, HONOR_INDICES
from mahjong.hand_calculating.yaku import Yaku
from mahjong.utils import is_chi


class Chanta(Yaku):
    """
    Every set must have at least one terminal or honour tile, and the pair must be of
    a terminal or honour tile. Must contain at least one sequence (123 or 789)
    """

    def __init__(self, yaku_id=None):
        super(Chanta, self).__init__(yaku_id)

    def set_attributes(self):
        self.tenhou_id = 23

        self.name = 'Chanta'
        self.english = 'Terminal Or Honor In Each Group'
        self.japanese = '混全帯么九'

        self.han_open = 1
        self.han_closed = 2

        self.is_yakuman = False

    def is_condition_met(self, hand, *args):
        def tile_in_indices(item_set, indices_array):
            for x in item_set:
                if x in indices_array:
                    return True
            return False

        honor_sets = 0
        terminal_sets = 0
        count_of_chi = 0
        for item in hand:
            if is_chi(item):
                count_of_chi += 1

            if tile_in_indices(item, TERMINAL_INDICES):
                terminal_sets += 1

            if tile_in_indices(item, HONOR_INDICES):
                honor_sets += 1

        if count_of_chi == 0:
            return False

        return terminal_sets + honor_sets == 5 and terminal_sets != 0 and honor_sets != 0



import copy

from mahjong.constants import EAST, FIVE_RED_MAN, FIVE_RED_PIN, FIVE_RED_SOU, TERMINAL_INDICES, CHUN


def is_aka_dora(tile, aka_enabled):
    """
    :param tile: int 136 tiles format
    :param aka_enabled: depends on table rules
    :return: boolean
    """

    if not aka_enabled:
        return False

    if tile in [FIVE_RED_MAN, FIVE_RED_PIN, FIVE_RED_SOU]:
        return True

    return False


def plus_dora(tile, dora_indicators):
    """
    :param tile: int 136 tiles format
    :param dora_indicators: array of 136 tiles format
    :return: int count of dora
    """
    tile_index = tile // 4
    dora_count = 0

    for dora in dora_indicators:
        dora //= 4

        # sou, pin, man
        if tile_index < EAST:

            # with indicator 9, dora will be 1
            if dora == 8:
                dora = -1
            elif dora == 17:
                dora = 8
            elif dora == 26:
                dora = 17

            if tile_index == dora + 1:
                dora_count += 1
        else:
            if dora < EAST:
                continue

            dora -= 9 * 3
            tile_index_temp = tile_index - 9 * 3

            # dora indicator is north
            if dora == 3:
                dora = -1

            # dora indicator is hatsu
            if dora == 6:
                dora = 3

            if tile_index_temp == dora + 1:
                dora_count += 1

    return dora_count


def is_chi(item):
    """
    :param item: array of tile 34 indices
    :return: boolean
    """
    if len(item) != 3:
        return False

    return item[0] == item[1] - 1 == item[2] - 2


def is_pon(item):
    """
    :param item: array of tile 34 indices
    :return: boolean
    """
    if len(item) != 3:
        return False

    return item[0] == item[1] == item[2]


def is_pair(item):
    """
    :param item: array of tile 34 indices
    :return: boolean
    """
    return len(item) == 2


def is_man(tile):
    """
    :param tile: 34 tile format
    :return: boolean
    """
    return tile <= 8


def is_pin(tile):
    """
    :param tile: 34 tile format
    :return: boolean
    """
    return 8 < tile <= 17


def is_sou(tile):
    """
    :param tile: 34 tile format
    :return: boolean
    """
    return 17 < tile <= 26


def is_honor(tile):
    """
    :param tile: 34 tile format
    :return: boolean
    """
    return tile >= 27


def is_terminal(tile):
    """
    :param tile: 34 tile format
    :return: boolean
    """
    return tile in TERMINAL_INDICES


def is_dora_indicator_for_terminal(tile):
    """
    :param tile: 34 tile format
    :return: boolean
    """
    return tile == 7 or tile == 8 or tile == 16 or tile == 17 or tile == 25 or tile == 26


def contains_terminals(hand_set):
    """
    :param hand_set: array of 34 tiles
    :return: boolean
    """
    return any([x in TERMINAL_INDICES for x in hand_set])


def simplify(tile):
    """
    :param tile: 34 tile format
    :return: tile: 0-8 presentation
    """
    return tile - 9 * (tile // 9)


def find_isolated_tile_indices(hand_34):
    """
    Tiles that don't have -1, 0 and +1 neighbors
    :param hand_34: array of tiles in 34 tile format
    :return: array of isolated tiles indices
    """
    isolated_indices = []

    for x in range(0, CHUN + 1):
        # for honor tiles we don't need to check nearby tiles
        if is_honor(x) and hand_34[x] == 0:
            isolated_indices.append(x)
        else:
            simplified = simplify(x)

            # 1 suit tile
            if simplified == 0:
                if hand_34[x] == 0 and hand_34[x + 1] == 0:
                    isolated_indices.append(x)
            # 9 suit tile
            elif simplified == 8:
                if hand_34[x] == 0 and hand_34[x - 1] == 0:
                    isolated_indices.append(x)
            # 2-8 tiles tiles
            else:
                if hand_34[x] == 0 and hand_34[x - 1] == 0 and hand_34[x + 1] == 0:
                    isolated_indices.append(x)

    return isolated_indices


def is_tile_strictly_isolated(hand_34, tile_34):
    """
    Tile is strictly isolated if it doesn't have -2, -1, 0, +1, +2 neighbors
    :param hand_34: array of tiles in 34 tile format
    :param tile_34: int
    :return: bool
    """
    hand_34 = copy.copy(hand_34)
    # we don't need to count target tile in the hand
    hand_34[tile_34] -= 1
    if hand_34[tile_34] < 0:
        hand_34[tile_34] = 0

    indices = []
    if is_honor(tile_34):
        return hand_34[tile_34] == 0
    else:
        simplified = simplify(tile_34)

        # 1 suit tile
        if simplified == 0:
            indices = [tile_34, tile_34 + 1, tile_34 + 2]
        # 2 suit tile
        elif simplified == 1:
            indices = [tile_34 - 1, tile_34, tile_34 + 1, tile_34 + 2]
        # 8 suit tile
        elif simplified == 7:
            indices = [tile_34 - 2, tile_34 - 1, tile_34, tile_34 + 1]
        # 9 suit tile
        elif simplified == 8:
            indices = [tile_34 - 2, tile_34 - 1, tile_34]
        # 3-7 tiles tiles
        else:
            indices = [tile_34 - 2, tile_34 - 1, tile_34, tile_34 + 1, tile_34 + 2]

    return all([hand_34[x] == 0 for x in indices])


def count_tiles_by_suits(tiles_34):
    """
    Separate tiles by suits and count them
    :param tiles_34: array of tiles to count
    :return: dict
    """
    suits = [
        {'count': 0, 'name': 'sou',   'function': is_sou},
        {'count': 0, 'name': 'man',   'function': is_man},
        {'count': 0, 'name': 'pin',   'function': is_pin},
        {'count': 0, 'name': 'honor', 'function': is_honor}
    ]

    for x in range(0, 34):
        tile = tiles_34[x]
        if not tile:
            continue

        for item in suits:
            if item['function'](x):
                item['count'] += tile

    return suits



# -*- coding: utf-8 -*-
from mahjong.constants import HAKU
from mahjong.hand_calculating.yaku import Yaku
from mahjong.utils import is_pon


class Haku(Yaku):
    """
    Pon of white dragons
    """

    def __init__(self, yaku_id=None):
        super(Haku, self).__init__(yaku_id)

    def set_attributes(self):
        self.tenhou_id = 18

        self.name = 'Yakuhai (haku)'
        self.english = 'White Dragon'
        self.japanese = '役牌(白)'

        self.han_open = 1
        self.han_closed = 1

        self.is_yakuman = False

    def is_condition_met(self, hand, *args):
        return len([x for x in hand if is_pon(x) and x[0] == HAKU]) == 1



# -*- coding: utf-8 -*-
from mahjong.hand_calculating.yaku import Yaku


class Houtei(Yaku):
    """
    Yaku situation
    """

    def __init__(self, yaku_id=None):
        super(Houtei, self).__init__(yaku_id)

    def set_attributes(self):
        self.tenhou_id = 6

        self.name = 'Houtei Raoyui'
        self.english = 'Win by last discard'
        self.japanese = '河底撈魚'

        self.han_open = 1
        self.han_closed = 1

        self.is_yakuman = False

    def is_condition_met(self, hand, *args):
        # was it here or not is controlling by superior code
        return True



# -*- coding: utf-8 -*-
from functools import reduce

from mahjong.constants import HONOR_INDICES, TERMINAL_INDICES
from mahjong.hand_calculating.yaku import Yaku


class Honroto(Yaku):
    """
    All tiles are terminals or honours
    """

    def __init__(self, yaku_id=None):
        super(Honroto, self).__init__(yaku_id)

    def set_attributes(self):
        self.tenhou_id = 31

        self.name = 'Honroutou'
        self.english = 'Terminals and Honors'
        self.japanese = '混老頭'

        self.han_open = 2
        self.han_closed = 2

        self.is_yakuman = False

    def is_condition_met(self, hand, *args):
        indices = reduce(lambda z, y: z + y, hand)
        result = HONOR_INDICES + TERMINAL_INDICES
        return all(x in result for x in indices)



# -*- coding: utf-8 -*-
from mahjong.hand_calculating.yaku import Yaku
from mahjong.utils import is_pon


class Suuankou(Yaku):
    """
    Four closed pon sets
    """

    def __init__(self, yaku_id=None):
        super(Suuankou, self).__init__(yaku_id)

    def set_attributes(self):
        self.tenhou_id = 41

        self.name = 'Suu ankou'
        self.english = 'Four Concealed Triplets'
        self.japanese = '四暗刻'

        self.han_open = None
        self.han_closed = 13

        self.is_yakuman = True

    def is_condition_met(self, hand, win_tile, is_tsumo):
        win_tile //= 4
        closed_hand = []
        for item in hand:
            # if we do the ron on syanpon wait our pon will be consider as open
            if is_pon(item) and win_tile in item and not is_tsumo:
                continue

            closed_hand.append(item)

        count_of_pon = len([i for i in closed_hand if is_pon(i)])
        return count_of_pon == 4



# -*- coding: utf-8 -*-
from mahjong.hand_calculating.yaku import Yaku
from mahjong.utils import is_pon, is_sou, is_pin, is_man, simplify


class SanshokuDoukou(Yaku):
    """
    Three pon sets consisting of the same numbers in all three suits
    """

    def __init__(self, yaku_id=None):
        super(SanshokuDoukou, self).__init__(yaku_id)

    def set_attributes(self):
        self.tenhou_id = 26

        self.name = 'Sanshoku Doukou'
        self.english = 'Three Colored Triplets'
        self.japanese = '三色同刻'

        self.han_open = 2
        self.han_closed = 2

        self.is_yakuman = False

    def is_condition_met(self, hand, *args):
        pon_sets = [i for i in hand if is_pon(i)]
        if len(pon_sets) < 3:
            return False

        sou_pon = []
        pin_pon = []
        man_pon = []
        for item in pon_sets:
            if is_sou(item[0]):
                sou_pon.append(item)
            elif is_pin(item[0]):
                pin_pon.append(item)
            elif is_man(item[0]):
                man_pon.append(item)

        for sou_item in sou_pon:
            for pin_item in pin_pon:
                for man_item in man_pon:
                    # cast tile indices to 1..9 representation
                    sou_item = [simplify(x) for x in sou_item]
                    pin_item = [simplify(x) for x in pin_item]
                    man_item = [simplify(x) for x in man_item]
                    if sou_item == pin_item == man_item:
                        return True
        return False



# -*- coding: utf-8 -*-
from mahjong.constants import EAST, SOUTH, WEST, NORTH
from mahjong.hand_calculating.yaku import Yaku
from mahjong.utils import is_pon, is_pair


class Shousuushii(Yaku):
    """
    The hand contains three sets of winds and a pair of the remaining wind
    """

    def __init__(self, yaku_id=None):
        super(Shousuushii, self).__init__(yaku_id)

    def set_attributes(self):
        self.tenhou_id = 50

        self.name = 'Shousuushii'
        self.english = 'Small Four Winds'
        self.japanese = '小四喜'

        self.han_open = 13
        self.han_closed = 13

        self.is_yakuman = True

    def is_condition_met(self, hand, *args):
        pon_sets = [x for x in hand if is_pon(x)]
        if len(pon_sets) < 3:
            return False

        count_of_wind_sets = 0
        wind_pair = 0
        winds = [EAST, SOUTH, WEST, NORTH]
        for item in hand:
            if is_pon(item) and item[0] in winds:
                count_of_wind_sets += 1

            if is_pair(item) and item[0] in winds:
                wind_pair += 1

        return count_of_wind_sets == 3 and wind_pair == 1



# -*- coding: utf-8 -*-
from mahjong.constants import SOUTH
from mahjong.hand_calculating.yaku import Yaku
from mahjong.utils import is_pon


class YakuhaiSouth(Yaku):
    """
    Pon of south winds
    """

    def __init__(self, yaku_id=None):
        super(YakuhaiSouth, self).__init__(yaku_id)

    def set_attributes(self):
        self.tenhou_id = 10

        self.name = 'Yakuhai (south)'
        self.english = 'South Round/Seat'
        self.japanese = '役牌(南)'

        self.han_open = 1
        self.han_closed = 1

        self.is_yakuman = False

    def is_condition_met(self, hand, player_wind, round_wind, *args):
        if len([x for x in hand if is_pon(x) and x[0] == player_wind]) == 1 and player_wind == SOUTH:
            return True

        if len([x for x in hand if is_pon(x) and x[0] == round_wind]) == 1 and round_wind == SOUTH:
            return True

        return False



from mahjong.tile import TilesConverter
from mahjong.meld import Meld
from mahjong.hand_calculating.hand_config import HandConfig, OptionalRules
from mahjong.hand_calculating.divider import HandDivider


class TestMixin(object):

    def _string_to_open_34_set(self, sou='', pin='', man='', honors=''):
        open_set = TilesConverter.string_to_136_array(sou=sou, pin=pin, man=man, honors=honors)
        open_set[0] //= 4
        open_set[1] //= 4
        open_set[2] //= 4
        return open_set

    def _string_to_34_tile(self, sou='', pin='', man='', honors=''):
        item = TilesConverter.string_to_136_array(sou=sou, pin=pin, man=man, honors=honors)
        item[0] //= 4
        return item[0]

    def _string_to_34_array(self, sou='', pin='', man='', honors=''):
        return TilesConverter.string_to_34_array(sou=sou, pin=pin, man=man, honors=honors)

    def _string_to_136_array(self, sou='', pin='', man='', honors=''):
        return TilesConverter.string_to_136_array(sou=sou, pin=pin, man=man, honors=honors)

    def _string_to_136_tile(self, sou='', pin='', man='', honors=''):
        return TilesConverter.string_to_136_array(sou=sou, pin=pin, man=man, honors=honors)[0]

    def _to_34_array(self, tiles):
        return TilesConverter.to_34_array(tiles)

    def _to_string(self, tiles_136):
        return TilesConverter.to_one_line_string(tiles_136)

    def _hand(self, tiles, hand_index=0):
        hand_divider = HandDivider()
        return hand_divider.divide_hand(tiles)[hand_index]

    def _make_meld(self, meld_type, is_open=True, man='', pin='', sou='', honors=''):
        tiles = self._string_to_136_array(man=man, pin=pin, sou=sou, honors=honors)
        meld = Meld(meld_type=meld_type, tiles=tiles, opened=is_open, called_tile=tiles[0], who=0)
        return meld

    def _make_hand_config(self,
                          is_tsumo=False,
                          is_riichi=False,
                          is_ippatsu=False,
                          is_rinshan=False,
                          is_chankan=False,
                          is_haitei=False,
                          is_houtei=False,
                          is_daburu_riichi=False,
                          is_nagashi_mangan=False,
                          is_tenhou=False,
                          is_renhou=False,
                          is_chiihou=False,
                          player_wind=None,
                          round_wind=None,
                          has_open_tanyao=False,
                          has_aka_dora=False,
                          disable_double_yakuman=False,
                          renhou_as_yakuman=False,
                          allow_daisharin=False,
                          allow_daisharin_other_suits=False):

        options = OptionalRules(
            has_open_tanyao=has_open_tanyao,
            has_aka_dora=has_aka_dora,
            has_double_yakuman=not disable_double_yakuman,
            renhou_as_yakuman=renhou_as_yakuman,
            has_daisharin=allow_daisharin,
            has_daisharin_other_suits=allow_daisharin_other_suits
        )
        return HandConfig(
            is_tsumo=is_tsumo,
            is_riichi=is_riichi,
            is_ippatsu=is_ippatsu,
            is_rinshan=is_rinshan,
            is_chankan=is_chankan,
            is_haitei=is_haitei,
            is_houtei=is_houtei,
            is_daburu_riichi=is_daburu_riichi,
            is_nagashi_mangan=is_nagashi_mangan,
            is_tenhou=is_tenhou,
            is_renhou=is_renhou,
            is_chiihou=is_chiihou,
            player_wind=player_wind,
            round_wind=round_wind,

            options=options
        )



# -*- coding: utf-8 -*-
from mahjong.hand_calculating.yaku import Yaku


class Renhou(Yaku):
    """
    Yaku situation
    """

    def __init__(self, yaku_id=None):
        super(Renhou, self).__init__(yaku_id)

    def set_attributes(self):
        self.tenhou_id = 36

        self.name = 'Renhou'
        self.english = 'Hand Of Man'
        self.japanese = '人和'

        self.han_open = None
        self.han_closed = 5

        self.is_yakuman = False

    def is_condition_met(self, hand, *args):
        # was it here or not is controlling by superior code
        return True



# -*- coding: utf-8 -*-
from mahjong.constants import FIVE_RED_MAN, FIVE_RED_PIN, FIVE_RED_SOU


class Tile(object):
    value = None
    is_tsumogiri = None

    def __init__(self, value, is_tsumogiri):
        self.value = value
        self.is_tsumogiri = is_tsumogiri


class TilesConverter(object):

    @staticmethod
    def to_one_line_string(tiles, print_aka_dora=False):
        """
        Convert 136 tiles array to the one line string
        Example of output with print_aka_dora=False: 1244579m3p57z
        Example of output with print_aka_dora=True:  1244079m3p57z
        """
        tiles = sorted(tiles)

        man = [t for t in tiles if t < 36]

        pin = [t for t in tiles if 36 <= t < 72]
        pin = [t - 36 for t in pin]

        sou = [t for t in tiles if 72 <= t < 108]
        sou = [t - 72 for t in sou]

        honors = [t for t in tiles if t >= 108]
        honors = [t - 108 for t in honors]

        def words(suits, red_five, suffix):
            return suits and ''.join(
                ["0" if i == red_five and print_aka_dora else str((i // 4) + 1) for i in suits]) + suffix or ''

        sou = words(sou, FIVE_RED_SOU - 72, 's')
        pin = words(pin, FIVE_RED_PIN - 36, 'p')
        man = words(man, FIVE_RED_MAN, 'm')
        honors = words(honors, -1 - 108, 'z')

        return man + pin + sou + honors

    @staticmethod
    def to_34_array(tiles):
        """
        Convert 136 array to the 34 tiles array
        """
        results = [0] * 34
        for tile in tiles:
            tile //= 4
            results[tile] += 1
        return results

    @staticmethod
    def to_136_array(tiles):
        """
        Convert 34 array to the 136 tiles array
        """
        temp = []
        results = []
        for x in range(0, 34):
            if tiles[x]:
                temp_value = [x * 4] * tiles[x]
                for tile in temp_value:
                    if tile in results:
                        count_of_tiles = len([x for x in temp if x == tile])
                        new_tile = tile + count_of_tiles
                        results.append(new_tile)

                        temp.append(tile)
                    else:
                        results.append(tile)
                        temp.append(tile)
        return results

    @staticmethod
    def string_to_136_array(sou=None, pin=None, man=None, honors=None, has_aka_dora=False):
        """
        Method to convert one line string tiles format to the 136 array.
        You can pass r or 0 instead of 5 for it to become a red five from
        that suit. To prevent old usage without red,
        has_aka_dora has to be True for this to do that.
        We need it to increase readability of our tests
        """
        def _split_string(string, offset, red=None):
            data = []
            temp = []

            if not string:
                return []

            for i in string:
                if (i == 'r' or i == '0') and has_aka_dora:
                    temp.append(red)
                    data.append(red)
                else:
                    tile = offset + (int(i) - 1) * 4
                    if tile == red and has_aka_dora:
                        # prevent non reds to become red
                        tile += 1
                    if tile in data:
                        count_of_tiles = len([x for x in temp if x == tile])
                        new_tile = tile + count_of_tiles
                        data.append(new_tile)

                        temp.append(tile)
                    else:
                        data.append(tile)
                        temp.append(tile)

            return data

        results = _split_string(man, 0, FIVE_RED_MAN)
        results += _split_string(pin, 36, FIVE_RED_PIN)
        results += _split_string(sou, 72, FIVE_RED_SOU)
        results += _split_string(honors, 108)

        return results

    @staticmethod
    def string_to_34_array(sou=None, pin=None, man=None, honors=None):
        """
        Method to convert one line string tiles format to the 34 array
        We need it to increase readability of our tests
        """
        results = TilesConverter.string_to_136_array(sou, pin, man, honors)
        results = TilesConverter.to_34_array(results)
        return results

    @staticmethod
    def find_34_tile_in_136_array(tile34, tiles):
        """
        Our shanten calculator will operate with 34 tiles format,
        after calculations we need to find calculated 34 tile
        in player's 136 tiles.

        For example we had 0 tile from 34 array
        in 136 array it can be present as 0, 1, 2, 3
        """
        if tile34 is None or tile34 > 33:
            return None

        tile = tile34 * 4

        possible_tiles = [tile] + [tile + i for i in range(1, 4)]

        found_tile = None
        for possible_tile in possible_tiles:
            if possible_tile in tiles:
                found_tile = possible_tile
                break

        return found_tile

    @staticmethod
    def one_line_string_to_136_array(string, has_aka_dora=False):
        """
        Method to convert one line string tiles format to the 136 array, like
        "123s456p789m11222z". 's' stands for sou, 'p' stands for pin,
        'm' stands for man and 'z' or 'h' stands for honor.
        You can pass r or 0 instead of 5 for it to become a red five from
        that suit. To prevent old usage without red,
        has_aka_dora has to be True for this to do that.
        """
        sou = ''
        pin = ''
        man = ''
        honors = ''

        split_start = 0

        for index, i in enumerate(string):
            if i == 'm':
                man += string[split_start: index]
                split_start = index + 1
            if i == 'p':
                pin += string[split_start: index]
                split_start = index + 1
            if i == 's':
                sou += string[split_start: index]
                split_start = index + 1
            if i == 'z' or i == 'h':
                honors += string[split_start: index]
                split_start = index + 1

        return TilesConverter.string_to_136_array(sou, pin, man, honors, has_aka_dora)

    @staticmethod
    def one_line_string_to_34_array(string, has_aka_dora=False):
        """
        Method to convert one line string tiles format to the 34 array, like
        "123s456p789m11222z". 's' stands for sou, 'p' stands for pin,
        'm' stands for man and 'z' or 'h' stands for honor.
        You can pass r or 0 instead of 5 for it to become a red five from
        that suit. To prevent old usage without red,
        has_aka_dora has to be True for this to do that.
        """
        results = TilesConverter.one_line_string_to_136_array(string, has_aka_dora)
        results = TilesConverter.to_34_array(results)
        return results



# -*- coding: utf-8 -*-


class HandResponse(object):
    cost = None
    han = None
    fu = None
    fu_details = None
    yaku = None
    error = None

    def __init__(self, cost=None, han=None, fu=None, yaku=None, error=None, fu_details=None):
        """
        :param cost: dict
        :param han: int
        :param fu: int
        :param yaku: list
        :param error: str
        :param fu_details: dict
        """
        self.cost = cost
        self.han = han
        self.fu = fu
        self.error = error

        if fu_details:
            self.fu_details = sorted(fu_details, key=lambda x: x['fu'], reverse=True)

        if yaku:
            self.yaku = sorted(yaku, key=lambda x: x.yaku_id)

    def __str__(self):
        if self.error:
            return self.error
        else:
            return '{} han, {} fu'.format(self.han, self.fu)



# -*- coding: utf-8 -*-
from mahjong.hand_calculating.yaku import Yaku


class DaburuKokushiMusou(Yaku):

    def __init__(self, yaku_id=None):
        super(DaburuKokushiMusou, self).__init__(yaku_id)

    def set_attributes(self):
        self.tenhou_id = 48

        self.name = 'Kokushi Musou Juusanmen Matchi'
        self.english = 'Thirteen Orphans 13-way wait'
        self.japanese = '国士無双十三面待ち'

        self.han_open = None
        self.han_closed = 26

        self.is_yakuman = True

    def is_condition_met(self, hand, *args):
        # was it here or not is controlling by superior code
        return True



# -*- coding: utf-8 -*-
from mahjong.hand_calculating.yaku import Yaku
from mahjong.utils import is_pon


class Toitoi(Yaku):
    """
    The hand consists of all pon sets (and of course a pair), no sequences.
    """

    def __init__(self, yaku_id=None):
        super(Toitoi, self).__init__(yaku_id)

    def set_attributes(self):
        self.tenhou_id = 28
        self.name = 'Toitoi'
        self.english = 'All Triplets'
        self.japanese = '対々和'

        self.han_open = 2
        self.han_closed = 2

        self.is_yakuman = False

    def is_condition_met(self, hand, *args):
        count_of_pon = len([i for i in hand if is_pon(i)])
        return count_of_pon == 4



# -*- coding: utf-8 -*-
from functools import reduce

from mahjong.constants import TERMINAL_INDICES, HONOR_INDICES
from mahjong.hand_calculating.yaku import Yaku


class Tanyao(Yaku):
    """
    Hand without 1, 9, dragons and winds
    """

    def __init__(self, yaku_id=None):
        super(Tanyao, self).__init__(yaku_id)

    def set_attributes(self):
        self.tenhou_id = 8

        self.name = 'Tanyao'
        self.english = 'All Simples'
        self.japanese = '断么九'

        self.han_open = 1
        self.han_closed = 1

        self.is_yakuman = False

    def is_condition_met(self, hand, *args):
        indices = reduce(lambda z, y: z + y, hand)
        result = TERMINAL_INDICES + HONOR_INDICES
        return not any(x in result for x in indices)



# -*- coding: utf-8 -*-
from mahjong.hand_calculating.yaku import Yaku


class SuuankouTanki(Yaku):

    def __init__(self, yaku_id=None):
        super(SuuankouTanki, self).__init__(yaku_id)

    def set_attributes(self):
        self.tenhou_id = 40

        self.name = 'Suu ankou tanki'
        self.english = 'Four Concealed Triplets Single Wait'
        self.japanese = '四暗刻単騎待ち'

        self.han_open = None
        self.han_closed = 26

        self.is_yakuman = True

    def is_condition_met(self, hand, *args):
        return True



from mahjong.hand_calculating.yaku_list.yakuman.chiihou import Chiihou
from mahjong.hand_calculating.yaku_list.yakuman.chinroto import Chinroutou
from mahjong.hand_calculating.yaku_list.yakuman.chuuren_poutou import ChuurenPoutou
from mahjong.hand_calculating.yaku_list.yakuman.daburu_chuuren_poutou import DaburuChuurenPoutou
from mahjong.hand_calculating.yaku_list.yakuman.daburu_kokushi import DaburuKokushiMusou
from mahjong.hand_calculating.yaku_list.yakuman.daisangen import Daisangen
from mahjong.hand_calculating.yaku_list.yakuman.daisharin import Daisharin
from mahjong.hand_calculating.yaku_list.yakuman.daisuushi import DaiSuushii
from mahjong.hand_calculating.yaku_list.yakuman.kokushi import KokushiMusou
from mahjong.hand_calculating.yaku_list.yakuman.ryuisou import Ryuuiisou
from mahjong.hand_calculating.yaku_list.yakuman.renhou_yakuman import RenhouYakuman
from mahjong.hand_calculating.yaku_list.yakuman.shosuushi import Shousuushii
from mahjong.hand_calculating.yaku_list.yakuman.suuankou import Suuankou
from mahjong.hand_calculating.yaku_list.yakuman.suuankou_tanki import SuuankouTanki
from mahjong.hand_calculating.yaku_list.yakuman.suukantsu import Suukantsu
from mahjong.hand_calculating.yaku_list.yakuman.tenhou import Tenhou
from mahjong.hand_calculating.yaku_list.yakuman.tsuisou import Tsuuiisou
